﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;

namespace W509CS
{
    class HexLine
    {
        public StreamWriter FileToWrite { get; set; }
        private StreamReader _fileHandle;
	    private int		_highAddress = 0;
	    private int		_lineCount = 0;
	    private bool 	_error =false;

        public HexLine() { }

        public HexLine(StreamReader sr)
        {
            _fileHandle = sr;
	    }

        public HexLine(StreamWriter sw)
        {
            FileToWrite = sw;
        }
	    //public virtual ~HexLine(){}

        private bool ByteToWord(byte[] data, out Int16[] Word)
        {
            if (data.Length % 2 != 0)
            {
                Word = null;
                return false;
            }
            Word = new Int16[data.Length / 2];
            for (int i = 0, j=0; i < data.Length-1; i += 2)
            {
                Word[j++] = (Int16)(data[i + 0] + (data[i + 1] << 8));
            }
            return true;
        }

        private bool WordToByte(Int16[] word, out byte[] data)
        {
            data = new Byte[word.Length * 2];
            for (int i = 0, j = 0; i < word.Length; i++)
            {
                Int16 aword = word[i];
                data[j++] = (byte)(aword & 0xff);
                data[j++] = (byte)((aword >> 8) & 0xff);
            }
            return true;
        }

        public int ReadData(out uint addr, out byte[] data)
        {
            // Read HexLine To byte data (8bit)
            // return   -1: error
            //          0: read data
            //          1: End
   
            addr = 0;
            if (_fileHandle == null)
            {
                data = null;
                return -1;
            }
            //addr = 0;
            //length = 0;
            int offset = 0, type = 0;
	        while ( GetLine(ref offset,ref type,out data)) 
            {
		        if (type == 0x00)
                {	// data record
			        addr = (uint)(offset + (_highAddress << 16));
			        //length = (byte)count;
			        return 0;
		        }
                else if (type == 0x01)
                { // end record
			        return 1;
		        }
                else if (type == 0x02)
                {
			        ;
		        }
                else if (type == 0x03)
                {
			        ;
		        }else if (type == 0x04)
                { // high address
			        _highAddress = data[0] << 8 | data[1];
		        }
                else if (type == 0x05)
                {
			        ;
		        }
	        }
	        return -1;
        }

        public int ReadWord(out uint addr, out Int16[] Word)
        {
            // Read Hexline To WordData(16bit)
            // return   -1: error
            //          0: read data
            //          1: End
            byte[] data;
            Word = null;    //??
            int status = ReadData(out addr, out data);
            if (status != 0)
            {
                Word = null;
                return status;
            }
            if (ByteToWord(data, out Word))
            {
                addr /= 2;
                return status;
            }
            else
                return -1;
        }

        public bool writeData(uint addr, byte[] data)
        {
            if (FileToWrite == null)
                return false;

	        if (_highAddress != (addr>>16)) 
            {
		        byte[] w = new byte[2];
		        _highAddress = (int)(addr>>16);
		        w[0] = (byte)(_highAddress >> 8);
		        w[1] = (byte)(_highAddress & 0xff);
		        putline(2, 0x0000, 0x04, w); // liner high address record
	        }
            int count = data.Length;
	        return putline(count, (int)(addr&0xffff), 0x00, data);
        }

        public bool writeWord(uint addr, Int16[] word)
        {
            byte[] data;
            WordToByte(word, out data);
            addr *= 2;
            return writeData(addr, data);

        }

        private bool GetLine(ref int offset,ref int type, out byte[] data)
        {
            // Parseで16進数変換用
            System.Globalization.NumberStyles HexStyles = System.Globalization.NumberStyles.HexNumber;

	        string Buf;
	        //while (fgets(buf, sizeof(buf), _fileHandle)){
            if(_fileHandle.Peek() >= 0)
            {
                Buf = _fileHandle.ReadLine();
		        _lineCount++;
		        if (Buf[0]==':')
                {
			        //char *p = buf + 1;
			        int dat;
			        byte sum;
			
                    try
                    {
                        //if (!gethexbyte(&count, p, 1))
                        //    goto error;
                        int p = 1;
                        int count = int.Parse(Buf.Substring(p,2), HexStyles);
                        data = new byte[count];
			            p += 2;
			
                        //if (!gethexbyte(&offset, p, 2))
                        //    goto error;
                        offset = int.Parse(Buf.Substring(p,4), HexStyles);
			            p+=4;
                        //if (!gethexbyte(&type, p, 1))
                        //    goto error;
                        type = int.Parse(Buf.Substring(p,2), HexStyles);
			            p+=2;
			
			            sum = (byte)(count + type + (offset >> 8) + (offset & 0xff));
			
			            for (int i=0;i<count;i++)
                        {
                            //if (!gethexbyte(&dat, p, 1))
                            //    goto error;
                            dat = int.Parse(Buf.Substring(p,2), HexStyles);
				            p+=2;
				            data[i] = (byte)dat;
				            sum += (byte)dat;
			            }
			
                        //if (!gethexbyte(&dat, p, 1))
                        //    goto error;
			            dat = int.Parse(Buf.Substring(p,2), HexStyles);
			            sum += (byte)dat;
			            if (sum != 0)
				            //goto error;
                            throw new Exception() ;
			
			            return true;
                    }
                    catch(Exception e)
                    {
                        data = null;
                        _error = true;
	                    return false;
                    }
		        }
	        }
            data = null;
            _error = true;
	        return false;
	
         //error:
         //   fprintf(stderr, "[ERROR] Wrong file format. line %d\n", _lineCount);
         //   _error = true;
         //   return false;
        }

        private bool putline(int count,int offset,int type,byte[] data)
        {
	        byte sum;
	
	        sum = (byte)(count + (offset >> 8) + (offset & 0xff) + type);
	
	        //fprintf(_fileHandle, ":%02X%04X%02X", count, offset, type);
	        //FileToWrite.Write(":{0:X2}{1:X4}{2:X2}",count, offset, type);
            string r = ":" + count.ToString("X2") + offset.ToString("X4") + type.ToString("X2");
	        for (int i=0; i<count; i++){
		        //fprintf(_fileHandle, "%02X", data[i]);
		        r += data[i].ToString("X2");
                sum += data[i];
	        }	
	        //fprintf(_fileHandle, "%02X\n", (unsigned char)-sum);
            r += ((byte)(-sum)).ToString("X2");
            FileToWrite.WriteLine(r);
	        return true;
        }


    }
}
