using System;
using Gtk;
using System.IO;
using System.Collections.Generic;
using System.Xml;	// need to add reference "System.Xml"


using W509CS;
using W509Mono;
using NDesk.Options;

public partial class MainWindow : Gtk.Window
{
	enum OptCommand
    {
        Read,
        Save,
        Verify,
        Erase,
        BlankCheck,
        DumpMemory
    }

	private string W509ConfigFile = "W509Mono.config";
	private string DeviceConfigFile = "DeviceConfig.dat";
	private Writer509 MyWriter;
	public string DeviceName;
	private string HexFileName;
//	private string PortName = "/dev/ttyACM0";
	private string PortName;
	private int BautRate = 38400;
	private int SerialTimeOut = 100;
	private bool help;
	public string CurDir;

	private XmlDocument xmlConfig;
	private XmlNode xmlDevice;
	private XmlNode xmlCurDir;
	private XmlNode xmlPort;
	
	
	public MainWindow (string[] Args) : base(Gtk.WindowType.Toplevel)
	{
		Build ();
		
	    //read from .Config
        //CurDir = Properties.Settings.Default.CurDir;
        //DeviceName = Properties.Settings.Default.Device;
        //openFileDialog1.InitialDirectory = CurDir;
		LoadW509Config();
		
	    //バージョンの取得
        System.Reflection.Assembly asm = System.Reflection.Assembly.GetExecutingAssembly();
        System.Version ver = asm.GetName().Version;
        AppendLog("W509Cs Ver." + ver.ToString());

	    // Set commandline options
        var p = new OptionSet () {
			{ "p|port=",    v => PortName = v },
            { "s|speed=",   v => BautRate = int.Parse(v) },
            { "t|timeout=", v => SerialTimeOut = int.Parse(v) },
            { "d|dvice=",   v => DeviceName = v },
            { "f|file=",    v => HexFileName = v },
            { "c|config=",  v => DeviceConfigFile = v },
            { "h|?|help",   v => help = v != null },
        };

        List<string> extra = p.Parse (Args);
        if (extra.Count > 0)
            AppendLog("Error: CommandLineOptions");

		
		if(LoadDevice())
			AppendLog(DeviceConfigFile + " loaded." );
		else
			AppendLog("Error:Device config file loaded." );
		

        MyWriter = new Writer509();
		MyWriter.AppendLog = this.AppendLog;
		MyWriter.BaudRate = BautRate;
		// MyWriter.TimeOut = SerialTimeOut;
		if(PortName != null)
			if(PortName.Length > 0){
				entryPort.Text = PortName;
				OnButtonPortClicked(null, null);
			}
		
		checkbuttonErase.Active = true;
		
		foreach(OptCommand Cmnd in Enum.GetValues(typeof(OptCommand)))
		{
			comboboxCmnd.AppendText(Cmnd.ToString());
		}
		comboboxCmnd.Active = 0;
		buttonLoad.Sensitive = false;
		buttonWrite.Sensitive = false;
		
		FileFilter hexFilter = new FileFilter();
		hexFilter.Name = "Hex File (*.hex)";
		//hexFilter.AddMimeType("hex");
		hexFilter.AddPattern("*.hex");
		filechooserbutton1.Filter = hexFilter;
		if(CurDir != null)
			if(CurDir.Length > 0)
				filechooserbutton1.SetCurrentFolder(CurDir);
		if(HexFileName != null)
			if(HexFileName.Length>0)
				filechooserbutton1.SetFilename(HexFileName);
		
	}

	protected void OnDeleteEvent (object sender, DeleteEventArgs a)
	{
	    //save to .Config
        //Properties.Settings.Default.CurDir = CurDir;
        //Properties.Settings.Default.Device = DeviceName;
        //Properties.Settings.Default.Save();

		if(xmlConfig != null){
			if(xmlDevice != null)
				xmlDevice.InnerText = comboboxDevice.ActiveText;
			if(xmlCurDir != null)
				xmlCurDir.InnerText = filechooserbutton1.CurrentFolder;
			if(xmlPort != null)
				xmlPort.InnerText = entryPort.Text;
			xmlConfig.Save(W509ConfigFile);
		}
		Application.Quit ();
		a.RetVal = true;
	}
	
	private bool LoadW509Config()
	{
		if(!File.Exists(W509ConfigFile))
			return false;
		
		xmlConfig = new XmlDocument(); 
		xmlConfig.Load(W509ConfigFile);
		//StreamReader srConfig = new StreamReader(W509ConfigFile);
		XmlNode xSettings = xmlConfig.SelectSingleNode("/configuration/userSettings/W509e16.Properties.Settings");
		if(xSettings.HasChildNodes){
			/*
			XmlNode aNode = xSettings.SelectSingleNode("[name='Device']");
			if(aNode != null){
				DeviceName = aNode.InnerText;}
			*/
			
			foreach(XmlNode aNode in xSettings.ChildNodes){
				XmlNode ValueNode = aNode.SelectSingleNode("value");
				if(ValueNode != null){
					XmlAttributeCollection atrbs = aNode.Attributes;
					var atrb = atrbs.Item(0);
					switch(atrb.Value){
						case "Device":
							xmlDevice = ValueNode;
							DeviceName = ValueNode.InnerText;
							break;
						case "CurDir":
							xmlCurDir = ValueNode;
							CurDir =  ValueNode.InnerText;
							break;
						case "SerialPort":
							xmlPort = ValueNode;
							PortName =  ValueNode.InnerText;
							break;
					}
				}
			}
			return true;
		}

		return false;
	}
	
	private bool LoadDevice()
	{
		StreamReader sr = new StreamReader(DeviceConfigFile);
		if(!PicDevice.LoadList(sr))
			return false;			

		//picDevices = PicDevice.DeviceList;
				int i=0;
		foreach(PicDevice aDevice in PicDevice.DeviceList)
		{
			comboboxDevice.AppendText(aDevice.Name);
			if(aDevice.Name == DeviceName)
				comboboxDevice.Active = i;
			i++;
		}
		if(DeviceName == null)
			comboboxDevice.Active = 0;
		//comboboxDevice.ActiveText = PicDevice.DeviceList[0].Name;
		return true;
	}

	protected virtual void OnButtonPortClicked (object sender, System.EventArgs e)
	{
		if(entryPort.Text.Length == 0)
		{
			AppendLog("Error: No Port name entered.");
			return;
		}
		else
			MyWriter.PortName = entryPort.Text;
		
		if(MyWriter.CheckVer())
		{
			AppendLog(MyWriter.StrVer + " found on " + entryPort.Text + "." );
		}
		else
		{
			AppendLog("No writer found on " + entryPort.Text + "." );
			return;
		}

	}

	protected virtual void OnFilechooserbutton1SelectionChanged (object sender, System.EventArgs e)
	{
		if(filechooserbutton1.Filename != null)
			if(filechooserbutton1.Filename.Length > 4)
				if(filechooserbutton1.Filename.EndsWith(".hex", StringComparison.OrdinalIgnoreCase)){
				   	buttonLoad.Sensitive = true;
					return;
				}
		
		buttonLoad.Sensitive = false;
	}

	
	protected virtual void OnButtonLoadClicked (object sender, System.EventArgs e)
	{
		if(filechooserbutton1.Filename.Length == 0)
		{
			AppendLog("Error: No file is selected." );
			return;
		}
			
        StreamReader SrHex = new StreamReader(filechooserbutton1.Filename);
        MyWriter.TargetDevice = PicDevice.SerchFromName(comboboxDevice.ActiveText);
        //DeviceName = MyWriter.TargetDevice.Name;

        if (MyWriter.LoadHex(SrHex))
        {
            AppendLog("Hex file loaded." );
			buttonWrite.Sensitive = true;
        }
        else
        {
            AppendLog("Error: Hex file not loaded." );
        }
        SrHex.Dispose();

	}
	
	
	protected virtual void OnButtonWriteClicked (object sender, System.EventArgs e)
	{
		if(entryPort.Text.Length == 0)
		{
			AppendLog("Error: No Port name entered.");
			return;
		}
		else
			MyWriter.PortName = entryPort.Text;
		
		if(MyWriter.CheckVer())
		{
			AppendLog(MyWriter.StrVer + " found on " + entryPort.Text + "." );
		}
		else
		{
			AppendLog("Error: No writer found on " + entryPort.Text + "." );
			return;
		}
		
		// Application.DoEvents();
		//while (Application.EventsPending ())
        //	Application.RunIteration ();

		
		if(MyWriter.TargetDevice.Name != comboboxDevice.ActiveText)
		{
			AppendLog("Error: Device mismutch." );
			return;
		}
        buttonWrite.Sensitive = false;
        if (MyWriter.WriteChip(checkbuttonErase.Active))
            AppendLog("Write success." );
        else
            AppendLog("Write failed." );

		buttonWrite.Sensitive = true;

	}
	
	protected virtual void OnButton3Clicked (object sender, System.EventArgs e)
	{
		/*
		if(MyWriter.TargetDevice == null)
		{
			MyWriter.TargetDevice = PicDevice.SerchFromName(comboboxDevice.ActiveText);
		}
		else if(MyWriter.TargetDevice.Name != comboboxDevice.ActiveText)
		{
			AppendLog("Error: Selected device mismatch." );
		}
		*/
		
		MyWriter.PortName = entryPort.Text;
		
	    OptCommand Cmnd = (OptCommand)Enum.Parse(typeof(OptCommand),comboboxCmnd.ActiveText);
        switch (Cmnd)
        {
            case OptCommand.Read:
                MyWriter.TargetDevice = PicDevice.SerchFromName(comboboxDevice.ActiveText);
                DeviceName = MyWriter.TargetDevice.Name;
                if(MyWriter.ReadFromChip())
                    AppendLog("Memory Read succeeded." );
                else
                    AppendLog("Memory Read Error." );
                break;
            case OptCommand.Save:
				SaveHexFile();	
                break;
			
            case OptCommand.Erase:
                MyWriter.TargetDevice = PicDevice.SerchFromName(comboboxDevice.ActiveText);
                DeviceName = MyWriter.TargetDevice.Name;
               if(MyWriter.EraseChip())
                    AppendLog("Chip erased." );
                else
                    AppendLog("Chip Erase Error." );
                break;
            case OptCommand.BlankCheck:
                MyWriter.TargetDevice = PicDevice.SerchFromName(comboboxDevice.ActiveText);
                DeviceName = MyWriter.TargetDevice.Name;
               if(MyWriter.VerifyChip(true))
                    AppendLog( "Chip is Blank." );
                else
                    AppendLog("Chip is not blank." );
                break;
            case OptCommand.Verify:
                if(MyWriter.VerifyChip(false))
                    AppendLog("Device Verify succeeded." );
                else
                    AppendLog("Device Verify Error." );
                break;
            case OptCommand.DumpMemory:
				String sDump = MyWriter.MemoryDump();
				if(sDump.Length > 0)
				{
					DialogDump aDialog = new DialogDump(sDump);
					aDialog.Run();
					aDialog.Destroy();
				}
                else
                    AppendLog("Nothing in the Buffer.");
                break;

        }
		
	}
	
	public void AppendLog(String strLog)
	{
		textview1.Buffer.Text += strLog + Environment.NewLine;
		Gtk.TextIter iter = textview1.Buffer.EndIter;
		textview1.ScrollToIter(iter,0,false,0,0);
		while (Application.EventsPending ())
        	Application.RunIteration ();
	}

	private void SaveHexFile()
	{
		//FileChooserDialog MyFCD = new FileChooserDialog("Save Hex File", this, Gtk.FileChooserAction.Open, Items);
		//MyFCD.Show();
		//String strmHex = MyFCD.Filename;
		
		FileFilter hexFilter = new FileFilter();
		hexFilter.Name = "Hex File (*.hex)";
		//hexFilter.AddMimeType("hex");
		hexFilter.AddPattern("*.hex");

		Gtk.FileChooserDialog fc=
			new Gtk.FileChooserDialog("Save Hex File",
		                            this,
		                            FileChooserAction.Save,
		                            "Cancel",ResponseType.Cancel,
		                            "Save",ResponseType.Apply);
		fc.Filter = hexFilter;
		fc.DoOverwriteConfirmation = true;
		
		if (fc.Run() == (int)ResponseType.Apply) 
		{
			StreamWriter sw = new StreamWriter(fc.Filename);
			if(MyWriter.SaveHex(sw))
			//sw.Close();
				AppendLog("Hex file saved.");
			else
				AppendLog("Error: Save hex file.");

		}
		fc.Destroy();
		return;
	}
	
	
	
	
	
	
}

