﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.IO.Ports;
using System.Threading;

namespace W509CS
{
    enum WAITTYPE
    {
        WAIT_PROG,
        WAIT_DATA,
        WAIT_CONF,
    };

    class PicCommand
    {
        //private const int TRYTIME = 4;
        //#define WRITER509 "W509"
        //#define WRITER648 "W648"

        protected const byte W_BASE      = 0x00;
        protected const byte W_SINGLE    = 0x01;
        protected const byte W_RECDAT    = 0x02;
        protected const byte W_INTERNAL  = 0x03;
        protected const byte W_SENDDATA  = 0x04;
        protected const byte W_EXTERNAL  = 0x05;
        protected const byte W_SINGLE_N  = 0x06;
        protected const byte W_RECDAT_N  = 0x07;
        protected const byte W_ADVANC_W  = 0x08;
        protected const byte W_SEND20    = 0x09;
        protected const byte W_RECV8     = 0x0a;
        protected const byte W_ADDRSET24 = 0x0b;
        protected const byte W_TABLE_R_N = 0x0c;
        protected const byte W_WDATA_7   = 0x0d;
        protected const byte W_EP_ADDR   = 0x0e;
        protected const byte W_EP_READ_N = 0x0f;
        protected const byte W_EP_WRITE  = 0x10;
        protected const byte W_EXP_BASE  = 0xff;

        protected const byte W2_VERSION = W_BASE;  // 0x00
        protected const byte W2_VDDUP    = 0x01;
        protected const byte W2_VDDDN    = 0x02;
        protected const byte W2_VDDUP2   = 0x03;
        protected const byte W2_VDDDN2   = 0x04;
        protected const byte W2_CHGBPS   = 0xff;

        protected const byte P_LOAD_CONFIG = 0x00;
        protected const byte P_LOAD_PRGMEM = 0x02;
        protected const byte P_LOAD_DATMEM = 0x03;
        protected const byte P_READ_PRGMEM = 0x04;
        protected const byte P_READ_DATMEM = 0x05;
        protected const byte P_INC_ADDR    = 0x06;
        protected const byte P_BEGIN_PROG  = 0x08;
        protected const byte P_ERASE_PGRMEM = 0x09;
        protected const byte P_ERASE_DATMEM = 0x0b;

        protected const byte PROG_VERIFY  = 0x20;

        // public var
        public string StrVer { get; protected set; }
        public bool Advance { get; set; }

        //public bool m_bCheck;
        //public bool m_bErase;
        //public bool m_bWrite;
        //public bool m_bVerify;
        //public bool m_bRead;
        //public bool m_bChgBps;
  
        //public bool m_bLVP;
        //public bool m_bQuite;
  
        public PicDevice TargetDevice { get; set;}
        public SerialPort Serial { get; set;}
  
        //public int m_nNewBps;
  
        //public StreamReader  m_pFHexI;
        //public StreamReader  m_pFHexO;
   
        // private var
        public PicMemory Memory {get; set;}
        private bool m_bExpWriter;

        // public method
        public PicCommand()
        {
            StrVer = String.Empty;
            Advance = true;
         }


        public PicCommand(PicDevice device, SerialPort serial)
        {
            TargetDevice = device;
            Serial = serial;
            StrVer = String.Empty;
            Advance = true;
        }


        // private function
        protected bool SendCmd(byte Cmnd)
        {   // Send writer 1 byte command to writer 509 and check the echo back.
            byte[] buf = new byte[1];
            buf[0] = Cmnd;
            Serial.Write(buf, 0, 1);
            try
            {
                buf[0] = (byte)Serial.ReadByte();
            }
            catch
            {
                return false;
            }
            if(buf[0] == Cmnd)
                return true;
            else
                return false;
        }

        protected bool RecvChar(out byte r)
        {
            try
            {
                r = (byte)Serial.ReadByte();
            }
            catch
            {
                r = 0;
                return false;
            }
            return true;
        }

        // public method
        //public bool PortOpen()
        //{
        //    try
        //    {
        //        Serial.Open();
        //    }
        //    catch
        //    {
        //        return false;
        //    }
        //    return true;
        //}

        //public bool PortClose()
        //{
        //    try
        //    {
        //        Serial.Close();
        //    }
        //    catch
        //    {
        //        return false;
        //    }
        //    return true;
        //}

        public bool CheckWriter()
        {
            m_bExpWriter=false;
            StrVer = String.Empty;

            if (!SendCmd(W_BASE))
                return false;

            if (!SendCmd(W2_VERSION))
                return false;
            
            byte[] buf = new byte[32];; 
            try
            {
                //buf = new byte[m_pSerial.BytesToRead];
                int n = Serial.BytesToRead;
                Serial.Read(buf, 0, n);
                StrVer = Encoding.ASCII.GetString(buf).Substring(0,n);
            }
            catch
            {
                return false;
            }
            //finally
            //{
            //    m_pSerial.Close();
            //}

            m_bExpWriter = true;
            return true;
        }   // End of CheckWriter()

        protected bool PowerON()
        {
            if (TargetDevice.PowerUp == PWTYPE.VddBeforeVpp)
            {
	            if (SendCmd(W_BASE) && SendCmd(W2_VDDUP))
	                return true;
            } 
            else 
            {
	            if (SendCmd(W_BASE) && SendCmd(W2_VDDUP2))
	                return true;
            }
            return false;
        }

        protected bool PowerDown()
        {
            if (TargetDevice.PowerUp == PWTYPE.VddBeforeVpp)
            {
	            if (SendCmd(W_BASE) && SendCmd(W2_VDDDN))
	                return true;
            } 
            else 
            {
	            if (SendCmd(W_BASE) && SendCmd(W2_VDDDN2))
	                return true;
            }
            //return PowerDownTryAgain();
            return false;
        }

//bool piccommand::PowerDownTryAgain(){
//  char buf[1024], tmp;
//  int n=0;

//  while (m_pSerial->RecvChar(&tmp));

//  for (int i=0;i<TRYTIME;i++){
//    m_pSerial->SendCmd(W_BASE);
//    n=m_pSerial->RecvChar(buf, sizeof(buf));
//    if (n)
//      break;
//  }
//  buf[n]=0;

//  if (strncmp(buf, WRITER509, strlen(WRITER509)))
//    return false;

//  if (m_bLVP)
//    m_pSerial->SendCmd(W_EXP_BASE);
//  else
//    m_pSerial->SendCmd(W_BASE);

//  if (!m_pSerial->SendCmd(W2_VDDDN))
//    return false;

//  return true;
//}

        public bool DeviceCheck()
        {
            if (TargetDevice.IDword == 0)
	            return true;    // device has no ID
  
            PowerON();
            if (!LoadConfig(0) || !Increment(6))
            {
	            //fprintf(stderr, "[ERROR] device check\n");
	            PowerDown();
	            return false;
            }
  
            Int16 ret;
            ReadProgMemory(out ret);
            PowerDown();
  
            //dbprintf(1, "DeviceCheck %04x -- %04x\n", m_pDevice->GetDeviceID(), ret);

            if (TargetDevice.IDword != (ret & 0x3fe0))
            {
	            //fprintf(stderr, "[ERROR]: device not match\n");
	            return false;
            }
  
            return true;
        }

        protected virtual bool BackupCalib()
        {
             Int16 tmp;

            if (TargetDevice.CalibAddrs.Length == 0)
                return false;

            if (Memory == null)
                Memory = new PicMemory(TargetDevice);

            PowerON();
            for (int i = 0; i < TargetDevice.CalibAddrs.Length; i++)
            {
                if (TargetDevice.CalibAddrs[i] > TargetDevice.IDAddr)
                {
                    LoadConfig(0);
                    Increment(TargetDevice.CalibAddrs[i] - TargetDevice.IDAddr);
                    if (!ReadProgMemory(out tmp))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] BackupCalib1\n");
                        return false;
                    }
                    //Memory.CalibData1 = (Int16)(tmp & TargetDevice.CalibMask1);
                    Memory.CalibDatas[i] = (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMasks[i])) & TargetDevice.ProgMask);
                }
                //else if (m_pDevice->GetCalibAddr1()>0x2000 > 0){
                else if (TargetDevice.CalibAddrs[i] > 0)
                {
                    //if (TargetDevice.ConfAddr < TargetDevice.IDAddr)
                    //    Increment();
                    Increment(TargetDevice.CalibAddrs[i]);
                    if (!ReadProgMemory(out tmp))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] BackupCalib2\n");
                        return false;
                    }
                    //Memory.CalibData1 = (Int16)(tmp & TargetDevice.CalibMask1);
                    Memory.CalibDatas[i] = (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMasks[i])) & TargetDevice.ProgMask);
                }
            }
            //if (TargetDevice.CalibAddr2 > 0x2000)
            //{
            //    LoadConfig(0);
            //    Increment(TargetDevice.CalibAddr2 - 0x2000);
            //    if (!ReadProgMemory(out tmp))
            //    {
            //        PowerDown();
            //        //fprintf(stderr, " [ERROR] BackupCalib3\n");
            //        return false;
            //    }
            //    //Memory.CalibData2 = (Int16)(tmp & TargetDevice.CalibMask2);
            //    Memory.CalibData2 = (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMask2)) & TargetDevice.PMMask);
            //}
            //else if (TargetDevice.CalibAddr2 > 0)
            //{
            //    if (TargetDevice.ConfAddr1 < 0x1000)
            //        Increment();
            //    Increment(TargetDevice.CalibAddr2);
            //    if (!ReadProgMemory(out tmp))
            //    {
            //        PowerDown();
            //        //fprintf(stderr, " [ERROR] BackupCalib4\n");
            //        return false;
            //    }
            //    //Memory.CalibData2 = (Int16)(tmp & TargetDevice.CalibMask2);
            //    Memory.CalibData2 = (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMask2)) & TargetDevice.PMMask);
            //}
  
            PowerDown();
            return true;
        }

        protected virtual bool RestoreCalib()
        {
            if (TargetDevice.CalibAddrs.Length == 0)
                return false;
            
            PowerON();
            for (int i = 0; i < TargetDevice.CalibAddrs.Length; i++)
            {

                if (TargetDevice.CalibAddrs[i] > TargetDevice.IDAddr)
                {
                    LoadConfig(0);
                    Increment(TargetDevice.CalibAddrs[i] - TargetDevice.IDAddr);
                    //if (!LoadProgMemory(Memory.CalibData1) || !SendBeginProg(WAITTYPE.WAIT_PROG)) //original
                    if (!LoadProgMemory(Memory.CalibDatas[i]) || !SendBeginProg(WAITTYPE.WAIT_CONF))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] RestoreCalib1\n");
                        return false;
                    }
                }
                else if (TargetDevice.CalibAddrs[i] > 0)
                {
                    //if (TargetDevice.ConfAddr < TargetDevice.IDAddr)
                    //    Increment();
                    Increment(TargetDevice.CalibAddrs[i]);
                    if (!LoadProgMemory(Memory.CalibDatas[i]) || !SendBeginProg(WAITTYPE.WAIT_PROG))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] RestoreCalib2\n");
                        return false;
                    }
                }
            }
            //if (TargetDevice.CalibAddr2 > 0x2000)
            //{
            //    LoadConfig(0);
            //    Increment(TargetDevice.CalibAddr2 - 0x2000);
            //    //if (!LoadProgMemory(Memory.CalibData2) || !SendBeginProg(WAITTYPE.WAIT_PROG)){    // original
            //    if (!LoadProgMemory(Memory.CalibData2) || !SendBeginProg(WAITTYPE.WAIT_CONF)){
            //        PowerDown();
            //        //fprintf(stderr, " [ERROR] RestoreCalib3\n");
            //        return false;
            //    }
            //}
            //else if (TargetDevice.CalibAddr2 > 0)
            //{
            //    if (TargetDevice.ConfAddr1 < 0x1000)
            //        Increment();
            //    Increment(TargetDevice.CalibAddr2);
            //    if (!LoadProgMemory(Memory.CalibData2) || !SendBeginProg(WAITTYPE.WAIT_PROG)){
            //        PowerDown();
            //        //fprintf(stderr, " [ERROR] BackupCalib4\n");
            //        return false;
            //    }
            //}
  
            PowerDown();
            return true;
        }

        protected bool ConfirmCalib()
        {
            Int16 tmp;

            if (TargetDevice.CalibAddrs.Length == 0)
                return false;

            PowerON();
            for (int i = 0; i < TargetDevice.CalibAddrs.Length; i++)
            {
                if (TargetDevice.CalibAddrs[i] > TargetDevice.IDAddr)
                {
                    LoadConfig(0);
                    Increment(TargetDevice.CalibAddrs[i] - TargetDevice.IDAddr);
                    if (!ReadProgMemory(out tmp))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] BackupCalib1\n");
                        return false;
                    }
                    //Memory.CalibData1 = (Int16)(tmp & TargetDevice.CalibMask1);
                    if(Memory.CalibDatas[i] != (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMasks[i])) & TargetDevice.ProgMask))
                        return false;
                }
                //else if (m_pDevice->GetCalibAddr1()>0x2000 > 0){
                else if (TargetDevice.CalibAddrs[i] > 0)
                {
                    //if (TargetDevice.ConfAddr < TargetDevice.IDAddr)
                    //    Increment();
                    Increment(TargetDevice.CalibAddrs[i]);
                    if (!ReadProgMemory(out tmp))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] BackupCalib2\n");
                        return false;
                    }
                    //Memory.CalibData1 = (Int16)(tmp & TargetDevice.CalibMask1);
                    if(Memory.CalibDatas[i] != (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMasks[i])) & TargetDevice.ProgMask))
                        return false;
                }
            }

            PowerDown();
            return true;
        }


        protected bool InternalTime(byte c, byte t)
        {
            if (SendCmd(W_INTERNAL) && 
	            SendCmd(c) &&
	            SendCmd(t) )
            {
	            return true;
            }
            //CommunicateErr();
            return false;
        }

        protected bool ExternalTime(byte c1, byte c2, byte t)
        {
            if (SendCmd(W_EXTERNAL) && 
	            SendCmd(c1) &&
	            SendCmd(c2) &&
	            SendCmd(t))
            {
	            return true;
            }
            //CommunicateErr();
            return false;
        }

        protected bool Increment()
        {
            if (SendCmd(W_SINGLE) && SendCmd(P_INC_ADDR))
	            return true;
            //CommunicateErr();
            return false;
        }

        protected bool Increment(int n)
        {
            if(n==1)
                return Increment();
            else if (SendCmd(W_SINGLE_N) && 
	            SendCmd(P_INC_ADDR) &&
	            SendCmd((byte)(n & 0xff)) &&
	            SendCmd((byte)((n >> 8) & 0xff)) )
            {
	            return true;
            }
            else
            {
                //CommunicateErr();
                return false;
            }
        }

        protected bool Setup(byte c)
        {
            if (SendCmd(W_SINGLE) && 
	            SendCmd(c) )
            {
	            return true;
            }
            //CommunicateErr();
            return false;
        }

        protected bool LoadConfig(Int16 n)
        {
            if (SendCmd(W_SENDDATA) && 
	            SendCmd(P_LOAD_CONFIG) &&
	            SendCmd((byte)(n & 0xff)) &&
	            SendCmd((byte)((n >> 8) & 0xff)) )
            {
	            return true;
            }
            //CommunicateErr();
            return false;
        }

        protected bool ReadProgMemory(out Int16 r)
        {
            byte c1, c2;
            r = 0;

            if (SendCmd(W_RECDAT) && 
	            SendCmd(P_READ_PRGMEM) &&
                RecvChar(out c1) && 
	            RecvChar(out c2) )
            {
	            r = (Int16)((c1 & 0xff) + c2*0x100);
	            return true;
            }
            //CommunicateErr();
            return false;
        }



        protected bool LoadProgMemory(Int16 d)
        {
            if (SendCmd(W_SENDDATA) && 
	            SendCmd(P_LOAD_PRGMEM) &&
	            SendCmd((byte)(d & 0xff)) && 
	            SendCmd((byte)((d >> 8) & 0xff)) )
            {
	            return true;
            }
            //CommunicateErr();
            return false;
        }

        protected bool AdvanceWrite4(Int16[] p, int iOffset, byte latchmask)
        {
            if (SendCmd(W_ADVANC_W))
            {
                int sum = 0;
                byte[] c = new byte[13];
                byte cx;

                // 前半4ワード
                switch (TargetDevice.ProgType)
                {
                    case PGTYPE.Internal08:
                        c[0] = P_LOAD_PRGMEM;
                        c[1] = P_BEGIN_PROG;
                        c[2] = 0x00;
                        break;
                    case PGTYPE.Internal18:
                        c[0] = P_LOAD_PRGMEM;
                        c[1] = 0x18;
                        c[2] = 0x00;
                        break;
                    case PGTYPE.External08:
                        c[0] = P_LOAD_PRGMEM;
                        c[1] = P_BEGIN_PROG;
                        c[2] = 0x0e;
                        break;
                    case PGTYPE.External18:
                        c[0] = P_LOAD_PRGMEM;
                        c[1] = 0x18;
                        c[2] = 0x17;
                        break;
                    case PGTYPE.External18A:
                        c[0] = P_LOAD_PRGMEM;
                        c[1] = 0x18;
                        c[2] = 0x0A;
                        break;
                    default:
                        return false;
                }

                //sum = (byte)(c[0] +c[1] + c[2]);
                //Serial.Write(c, 0, 3);
                c[3] = (byte)TargetDevice.ProgWait;
                //m_pSerial->SendChar(m_pDevice->GetProgWait());
                //if (TargetDevice.ProgLatch == 8)
                //    c[4] = 0xFF;
                //else
                //    c[4] = (byte)(TargetDevice.ProgLatch - 1);
                c[4] = latchmask;
                //m_pSerial->SendChar(m_pDevice->GetProgLatch()-1);

                for (int i = 0, j = 5; i < 4; i++)
                {
                    c[j++] = (byte)(p[iOffset + i] & 0xff);
                    //m_pSerial->SendChar(p[i] & 0xff);
                    c[j++] = (byte)((p[iOffset + i] >> 8) & 0xff);
                    //m_pSerial->SendChar((p[i]>>8) & 0xff);
                }
                Serial.Write(c, 0, 13);
                RecvChar(out cx);

                //dbprintf(15, "ADV Chksum1: %02x %02x\n", cx & 0xff, sum & 0xff);
                for (int i = 0; i < 13; i++)
                {
                    sum += c[i];
                }

                sum -= cx;
                if ((sum & 0xff) == 0)
                    return true;
            }
            //fprintf(stderr, "[ERROR] advanced checksum error1\n");
            return false;
        }

        protected bool LoadProgMemory4N(Int16[] p, int iOffset, int N)
        {   //4xNワード　ロード using advance write 4
            byte LatchMask = (byte)(TargetDevice.ProgLatch - 1);
            for (int i = 0; i < N; i++)
            {
                if (TargetDevice.ProgLatch > 4)
                {
                    if (((i * 4) % TargetDevice.ProgLatch) == 0)
                        LatchMask = 0xFF;
                    else
                        LatchMask = 3;
                }
                if (!AdvanceWrite4(p, iOffset + i * 4, LatchMask))
                    return false;
            }
            return true;
        }

        protected bool LoadProgMemory8(Int16[] p, int iOffset)
        {   //8ワード　ロード
            if (SendCmd(W_ADVANC_W))
            {
	            int sum = 0;
	            byte[] c = new byte[13]; 
                byte cx;

                // 前半4ワード
	            switch(TargetDevice.ProgType)
                {
	            case PGTYPE.Internal08:
	                c[0] = P_LOAD_PRGMEM;
	                c[1] = P_BEGIN_PROG;
	                c[2] = 0x00;
	                break;
	            case PGTYPE.Internal18:
	                c[0] = P_LOAD_PRGMEM;
	                c[1] = 0x18;
	                c[2]=0x00;
	                break;
	            case PGTYPE.External08:
	                c[0] = P_LOAD_PRGMEM;
	                c[1] = P_BEGIN_PROG;
	                c[2] = 0x0e;
	                break;
	            case PGTYPE.External18:
	                c[0] = P_LOAD_PRGMEM;
	                c[1] = 0x18;
	                c[2] = 0x17;
	                break;
	            default:
	                return false;
	            }

	            //sum = (byte)(c[0] +c[1] + c[2]);
                //Serial.Write(c, 0, 3);
	            c[3] = (byte)TargetDevice.ProgWait;
	            //m_pSerial->SendChar(m_pDevice->GetProgWait());
                if(TargetDevice.ProgLatch == 8)
                    c[4] = 0xFF;
                else
	                c[4] = (byte)(TargetDevice.ProgLatch - 1);
	            //m_pSerial->SendChar(m_pDevice->GetProgLatch()-1);

	            for (int i=0, j=5; i<4; i++)
                {
	              c[j++] = (byte)(p[iOffset + i] & 0xff);
	              //m_pSerial->SendChar(p[i] & 0xff);
	              c[j++] = (byte)( (p[iOffset + i] >> 8) & 0xff);
	              //m_pSerial->SendChar((p[i]>>8) & 0xff);
	            }
                Serial.Write(c, 0, 13);
	            RecvChar(out cx);

	            //dbprintf(15, "ADV Chksum1: %02x %02x\n", cx & 0xff, sum & 0xff);
                for(int i=0; i<13; i++)
                {
                    sum += c[i];
                }

	            sum -= cx;
	            if ((sum & 0xff) != 0)
	                //fprintf(stderr, "[ERROR] advanced checksum error1\n");
                    return false;

                // 後半４ワード
	            if (SendCmd(W_ADVANC_W))
                {

                    //sum=c1+c2+c3;
                    //m_pSerial->SendChar(c1);
                    //m_pSerial->SendChar(c2);
                    //m_pSerial->SendChar(c3);
                    //sum += m_pDevice->GetProgWait();
                    //m_pSerial->SendChar(m_pDevice->GetProgWait());
                    //if (m_pDevice->GetProgLatch()==8){
                    //sum += 3;
                    //m_pSerial->SendChar(3);
                    //} else {
                    //sum += m_pDevice->GetProgLatch()-1;
                    //m_pSerial->SendChar(m_pDevice->GetProgLatch()-1);
                    //}
                    if(TargetDevice.ProgLatch == 8)
                        c[4] = 3;
                    else
	                    c[4] = (byte)(TargetDevice.ProgLatch - 1);

                    //for (int i=4;i<8;i++){
                    //sum += p[i];
                    //m_pSerial->SendChar(p[i] & 0xff);
                    //sum += p[i] >> 8;
                    //m_pSerial->SendChar((p[i]>>8) & 0xff);
                    //}
	                for (int i=4, j=5; i<8; i++)
                    {
	                  c[j++] = (byte)(p[iOffset + i] & 0xff);
	                  //m_pSerial->SendChar(p[i] & 0xff);
	                  c[j++] = (byte)( (p[iOffset + i] >> 8) & 0xff);
	                  //m_pSerial->SendChar((p[i]>>8) & 0xff);
	                }
                    Serial.Write(c, 0, 13);
	                RecvChar(out cx);
                    for(int i=0; i<13; i++)
                    {
                        sum += c[i];
                    }

	                //dbprintf(15, "ADV Chksum: %02x %02x\n", cx & 0xff, sum & 0xff);

	                sum -= cx;
	                if ((sum & 0xff) != 0)
		                //fprintf(stderr, "[ERROR] advanced checksum error2\n");
	                    return false;
	                return true;
	        }
          }
          return false;
        }

        protected bool ReadDataMemory(out Int16 r)
        {
            byte c1, c2;

            if (SendCmd(W_RECDAT) && 
	            SendCmd(P_READ_DATMEM) &&
	            RecvChar(out c1) && 
	            RecvChar(out c2) )
            {
	            r = (Int16)((c1 & 0xff) + c2 * 0x100);
	            return true;
            }
           
            r = 0;
            //CommunicateErr();
            return false;
        }


        protected bool BulkeraseF509()
        {
            return InternalTime(P_ERASE_PGRMEM, (byte)TargetDevice.EraseWait);
        }

        protected bool BulkeraseF675()
        {
            if (LoadConfig(0) &&
	                InternalTime(P_ERASE_PGRMEM, (byte)TargetDevice.EraseWait) && 
	                InternalTime(P_ERASE_DATMEM, (byte)TargetDevice.EraseWait) )
	            return true;
            else
                return false;
        }

        protected bool BulkeraseF683()
        {
            if (InternalTime(P_ERASE_PGRMEM, (byte)TargetDevice.EraseWait) &&
	                LoadConfig(0) &&
	                InternalTime(P_ERASE_PGRMEM, (byte)TargetDevice.EraseWait) && 
	                InternalTime(P_ERASE_DATMEM, (byte)TargetDevice.EraseWait) )
	            return true;
            else
                return false;
        }

        protected bool BulkeraseF84()
        {
            if (LoadConfig(0x3fff) &&
	                Increment(0x07) &&
	                Setup(0x01) &&
	                Setup(0x07) &&
	                InternalTime(P_BEGIN_PROG, (byte)TargetDevice.EraseWait) &&
	                Setup(0x01) &&
	                Setup(0x07) &&
	                Setup(P_ERASE_DATMEM) &&
	                InternalTime(0x18, (byte)TargetDevice.EraseWait))
	            return true;
            else
                return false;
        }

        protected bool BulkeraseF648A()
        {
            if (LoadConfig(0x3fff) &&
	                InternalTime(P_ERASE_PGRMEM, (byte)TargetDevice.EraseWait) && 
	                InternalTime(P_ERASE_DATMEM, (byte)TargetDevice.EraseWait))
	            return true;
            else
                return false;
        }

        protected bool ChiperaseF88()
        {
            if (LoadConfig(0x3fff) &&
	                InternalTime(0x1f, (byte)TargetDevice.EraseWait))
	            return true;
            else
                return false;
        }

        protected bool ChiperaseF77()
        {
            if (LoadConfig(0x3fff) &&
	                InternalTime(P_ERASE_PGRMEM, (byte)TargetDevice.EraseWait))
	            return true;
            else
                return false;
        }


        public virtual bool ChipErase()
        {
            //if (!m_bQuite)
            //    fprintf(stderr, "Chip Erase...");

            if (TargetDevice.CalibAddrs.Length > 0)
            {
                if (TargetDevice.CalibAddrs[0] != 0)
                    if (!BackupCalib())
                        return false;
            }
            else
                return false;
            PowerON();
  
            bool ret=false;
            switch(TargetDevice.EraseType)
            {
                case ESTYPE.BulkEraseF509:
	                ret=BulkeraseF509();
	                break;
                case ESTYPE.BulkEraseF675:
	                ret=BulkeraseF675();
	                break;
                case ESTYPE.BulkEraseF683:
	                ret=BulkeraseF683();
	                break;
                case ESTYPE.BulkEraseF84:
	                ret=BulkeraseF84();
	                break;
                case ESTYPE.BulkEraseF648A:
	                ret=BulkeraseF648A();
	                break;
                case ESTYPE.ChipEraseF88:
	                ret=ChiperaseF88();
	                break;
                case ESTYPE.ChipEraseF77:
	                ret=ChiperaseF77();
	                break;
                default:
	                //fprintf(stderr, " unknown erase type\n");
	                break;
            }

            PowerDown();

            if (TargetDevice.CalBackUp && TargetDevice.CalibAddrs.Length > 0)
            {
                if (TargetDevice.CalibAddrs[0] != 0)
                    ret = ret && RestoreCalib();
            }
            else
            {
                ret = ret && ConfirmCalib();
            }
            //if (!m_bQuite)
            //fprintf(stderr, "Done\n");
  
            return ret;
        }

        public virtual bool ReadBuffer(PicMemory a_memory)
        {
            if (Memory == null)
                return false;

            PowerON();
  
            int nSize, nMask;
            Int16[] Buffer;
            byte h,l;
  
            //if (!m_bQuite)
	        //    fprintf(stderr, "Program Area Reading");
  
            nSize = TargetDevice.ProgSize;
            nMask = TargetDevice.ProgMask;
            Buffer = a_memory.ProgMem;
            if (!(  SendCmd(W_RECDAT_N) && 
		            SendCmd(P_READ_PRGMEM) && 
		            SendCmd((byte)(nSize & 0xff)) && 
		            SendCmd((byte)((nSize >> 8) & 0xff)) ) )
            {
	            PowerDown();
	            return false;
            }
  
            for (int i=0; i<nSize; i++)
            {
	            if (RecvChar(out l) && RecvChar(out h)) 
                {
	                Buffer[i]=(Int16)((h*0x100 + (l & 0xff)) & nMask);
	            } 
                else 
                {
	                //fprintf(stderr, "Error\n");
	                PowerDown();
	                return false;
	            }
	            //if (!m_bQuite && i%0x20 == 0)
	            //    fprintf(stderr, ".");
            }
  
            //if (!m_bQuite)
            //fprintf(stderr, "Done\n");
  
            if (RecvChar(out l))//Error if extra data recieved 
            {
	            //fprintf(stderr, "communication error [%02x]\n", l);
	            PowerDown();
	            return false;
            }
  
            Int16 nData;
            if (TargetDevice.DataSize > 0)
            {
            //if (!m_bQuite)
            //    fprintf(stderr, "Data Area Reading");
	
	        nSize = TargetDevice.DataSize;
	        nMask = TargetDevice.DataMask;
	        Buffer = a_memory.DataMem;
	
	        for (int i=0;i<nSize;i++)
            {
	            if (!ReadDataMemory(out nData))
                {
		            PowerDown();
		            return false;
	            }
	            Buffer[i] = (Int16)(nData & nMask);
	            Increment();
	  
	            //if (!m_bQuite && i%0x10 == 0)
		        //  fprintf(stderr, ".");
	        }
	
            //if (!m_bQuite)
            //    fprintf(stderr, "Done\n");
            }
  
            nSize = TargetDevice.ConfDspSize;
            nMask = TargetDevice.ProgMask;//********************
            Buffer=a_memory.ConfMem;
  
            //if (!m_bQuite)
            //    fprintf(stderr, "Config Area Reading");
  
            if (!LoadConfig(0))
            {
	            PowerDown();
	            return false;
            }
  
            for (int i=0; i<nSize; i++)
            {
	            if (!ReadProgMemory(out nData))
                {
	                PowerDown();
	                return false;
	            }
	            Buffer[i] = (Int16)(nData & nMask);
	            Increment();
                //if (!m_bQuite)
	            //    fprintf(stderr, ".");
            }
  
            //if (!m_bQuite)
            //    fprintf(stderr, "Done\n");
  
            PowerDown();
            return true;
        }

        public bool ReadBuffer()
        {
            return ReadBuffer(Memory);
        }

        protected bool SendBeginProg(WAITTYPE eWait)
        {
            int nWait;

            switch (eWait)
            {
                case WAITTYPE.WAIT_PROG:
	                nWait=TargetDevice.ProgWait;
	                break;
                case WAITTYPE.WAIT_DATA:
	                nWait=TargetDevice.ProgWaitD;
	                break;
                case WAITTYPE.WAIT_CONF:
	                nWait=TargetDevice.ProgWaitC;
	                break;
                default:
	                //fprintf(stderr, "[*] Param Error piccommand::SendBeginProg()\n");
	                return false;
            }

            bool bRet=false;
            switch(TargetDevice.ProgType)
            {
                case PGTYPE.Internal08:
	                bRet = InternalTime(P_BEGIN_PROG, (byte)nWait);
	                break;
                case PGTYPE.Internal18:
	                bRet = InternalTime(0x18, (byte)nWait);
	                break;
                case PGTYPE.External08:
	                bRet = ExternalTime(P_BEGIN_PROG, 0x0e, (byte)nWait);
	                break;
                case PGTYPE.External18:
	                bRet = ExternalTime(0x18, 0x17, (byte)nWait);
	                break;
                case PGTYPE.External18A:
                    bRet = ExternalTime(0x18, 0x0A, (byte)nWait);   // added by matsu
                    break;
                default:
	                break;
            }
            //if (!bRet)
	        //fprintf (stderr, "[ERROR]: Program2\n");
  
            return bRet;
        }

        public virtual bool WriteBuffer()
        {
            if (Memory == null)
                return false;

            PowerON();
  
            int nSize, nMask;
            Int16[] pBuffer;
  
            //if (!m_bQuite)
            //    fprintf(stderr, "Program Area Writing");
  
            nSize = TargetDevice.ProgSize;
            nMask = TargetDevice.ProgMask;
            pBuffer = Memory.ProgMem;
            if(TargetDevice.CalibAddrs.Length > 0)
                if ((TargetDevice.CalibAddrs[0] > 0) && (TargetDevice.CalibAddrs[0] < TargetDevice.IDAddr))
	                nSize--;
 
            bool bAdv15 = Advance;
            int nIncrement = 0;
            int LatSize = TargetDevice.ProgLatch <= 4 ? 4 : TargetDevice.ProgLatch;

            for (int i=0;i<nSize;)
            {
                //if (!m_bQuite && i%0x20 == 0)
                //{
                //    fprintf(stderr, ".");
                //}

	            if (pBuffer[i] != nMask)
                {
	                if (nIncrement>0)
                    {
                        //if (nIncrement==1)
                        //    Increment();
                        //else
		                Increment(nIncrement);
		                nIncrement=0;
	                }

	                // I don't know why condition i%8 is necessary but it needs.
                    if (bAdv15 && i % LatSize == 0 && i < nSize - LatSize)
                    {
		                //if ( LoadProgMemory8(pBuffer+i))
                        if (LoadProgMemory4N(pBuffer, i, LatSize/4))
                        {
                            i += LatSize;
		                    continue;
		                }
		                bAdv15=false;
	                }

	                if (LoadProgMemory(pBuffer[i]))
                    {
		                if (!SendBeginProg(WAITTYPE.WAIT_PROG))
                        {
		                    PowerDown();
		                    return false;
		                }
	                }
                    else 
                    {
		                PowerDown();
		                return false;
	                }
	            }
	            i++;
	            nIncrement++;
            }

            PowerDown();
            //if (!m_bQuite)
            //    fprintf(stderr, "Done\n");
  
            if (TargetDevice.DataSize > 0)
            {
                PowerON();
                //if (!m_bQuite)
                //    fprintf(stderr, "Data Area Writing");
	
	            nSize = TargetDevice.DataSize;
	            nMask = TargetDevice.DataMask;
	            pBuffer = Memory.DataMem;
	
	            nIncrement = 0;

	            for (int i=0;i<nSize;i++)
                {
	                if (pBuffer[i] != nMask)
                    {
		                if (nIncrement > 0)
                        {
                            //if (nIncrement==1)
                            //Increment();
                            //else
			                Increment(nIncrement);
		                    nIncrement=0;
		                }

		                if (!(SendCmd(W_SENDDATA) && 
			                    SendCmd(P_LOAD_DATMEM) &&
			                    SendCmd((byte)(pBuffer[i] & 0xff)) &&
			                    SendCmd((byte)((pBuffer[i] >> 8) & 0xff))) )
                        {
		                    //fprintf (stderr, "[ERROR]: Program3\n");
		                    PowerDown();
		                    return false;
		                }
		                if (!SendBeginProg(WAITTYPE.WAIT_DATA))
                        {
		                    PowerDown();
		                    return false;
		                }
	                }
	                nIncrement++;
                    //if (!m_bQuite && i%0x10 == 0)
                    //fprintf(stderr, ".");
	            }
                PowerDown();
            }
  
            //if (!m_bQuite)
            //    fprintf(stderr, "Done\n");
            Thread.Sleep(10);   //test for PIC16F627A ***************
            PowerON();

            nSize = TargetDevice.ConfDspSize;
            //nMask = TargetDevice.ConfMask;
            pBuffer = Memory.ConfMem;
  
            //if (!m_bQuite)
            //    fprintf(stderr, "Config Area Writing");
  
            LoadConfig(TargetDevice.ProgMask);
            for (int i=0;i<4;i++)
            {
	            //if (!LoadProgMemory((Int16)(pBuffer[i] & 0x0f)))
                Int16 Uid = (Int16)(pBuffer[i] & TargetDevice.UIDMask);
                if (Uid != TargetDevice.UIDMask)
                {
                    if (!LoadProgMemory(Uid))
                    {
                        //fprintf (stderr, "[ERROR]: Program1\n");
                        PowerDown();
                        return false;
                    }
                    //if (!SendBeginProg(WAITTYPE.WAIT_PROG))   // original
                    if (!SendBeginProg(WAITTYPE.WAIT_CONF))
                    {
                        PowerDown();
                        return false;
                    }
                }
	            Increment();
            }
            //if (!m_bQuite)
            //    fprintf(stderr, ".");
            int iConf = TargetDevice.ConfAddr[0] - TargetDevice.IDAddr;
            Increment(iConf-4);

            for (int i = 0; i < TargetDevice.ConfAddr.Length; i++)
            {
                Int16 nConf;
                if (!ReadProgMemory(out nConf))
                {
                    PowerDown();
                    return false;
                }

                nConf = (Int16)((pBuffer[iConf] & TargetDevice.ConfMask[i]) | (nConf & ~TargetDevice.ConfMask[i]));
                if (!LoadProgMemory(nConf))
                {
                    //fprintf (stderr, "[ERROR]: Program4\n");
                    PowerDown();
                    return false;
                }
                if (!SendBeginProg(WAITTYPE.WAIT_CONF))
                {
                    PowerDown();
                    return false;
                }

                Increment();
                iConf++;
            }
  
            PowerDown();
            return true;
        }

        public int VerifyBuffer()
        {
            // returns
            //  -1: error
            //  0: OK
            //  0b0001: NG in program memory
            //  0b0010: NG in Data memory
            //  0b0100: NG in User ID
            //  0b1000: NG in Configuration

            int nSize, nMask;
            Int16[] rBuffer, wBuffer;
            PicMemory MemoryRead = new PicMemory(TargetDevice);
            if(!ReadBuffer(MemoryRead))
                return -1;

            //if (!m_bQuite)
            //    fprintf(stderr, "Program Area Verifing");

            int ret = 0;
            nSize = TargetDevice.ProgSize;
            nMask = TargetDevice.ProgMask;
            wBuffer = Memory.ProgMem;
            rBuffer = MemoryRead.ProgMem;

            if (TargetDevice.CalibAddrs.Length > 0)
                if (TargetDevice.CalibAddrs[0] > 0 && TargetDevice.CalibAddrs[0] < TargetDevice.IDAddr)
                    nSize--;

            for (int i=0; i < nSize; i++)
            {
                if(wBuffer[i] != rBuffer[i])
                {
                    ret += 1;
                }
            }
 
            //if (!m_bQuite)
            //    fprintf(stderr, "OK\n");

            if (TargetDevice.DataSize > 0)
            {
                //if (!m_bQuite)
                //    fprintf(stderr, "Data Area Verifing");

                nSize = TargetDevice.DataSize;
                nMask = TargetDevice.DataMask;
                wBuffer = Memory.DataMem;
                rBuffer = MemoryRead.DataMem;

                for (int i = 0; i < nSize; i++)
                {
                    if(wBuffer[i] != rBuffer[i])
                    {
                        ret += 2; ;
                    }
                }
            }

            nSize = TargetDevice.ConfDspSize;
            wBuffer = Memory.ConfMem;
            rBuffer = MemoryRead.ConfMem;
            // UserID
             for (int i=0;i<4;i++)
            {
                if ((rBuffer[i] & TargetDevice.UIDMask) != (wBuffer[i] & TargetDevice.UIDMask))
                {
                    ret += 4; ;
	            }
            }

            // Configuration word
             for (int i = 0; i < TargetDevice.ConfAddr.Length; i++)
             {
                 if (TargetDevice.ConfAddr[i] > 0x2000)
                 {
                     int iConf = TargetDevice.ConfAddr[i] - TargetDevice.IDAddr;
                     if ((rBuffer[iConf] & TargetDevice.ConfMask[i]) != (wBuffer[iConf] & TargetDevice.ConfMask[i]))
                     {
                         ret |= 8;
                     }
                 }
             }

            return ret;
        }
        

    }
}
