﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO.Ports;
using System.Threading;

namespace W509CS
{
    class PicCommandBL : PicCommand
    {
        public PicCommandBL(PicDevice device, SerialPort serial) : base(device, serial)
        {
            //TargetDevice = device;
            //Serial = serial;
            //StrVer = String.Empty;
            //Advance = true;
        }

        protected override bool BackupCalib()
        {
            Int16 tmp;

            if (TargetDevice.CalibAddrs.Length == 0)
                return false;

            if (Memory == null)
                Memory = new PicMemory(TargetDevice);

            PowerON();

            int addr = -1;// Reset vector = End of address
            for (int i = 0; i < TargetDevice.CalibAddrs.Length; i++)
            {
                int inc = TargetDevice.CalibAddrs[i] - addr;
                if (inc > 0)
                {
                    Increment(inc);
                    addr += inc;
                    if (!ReadProgMemory(out tmp))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] BackupCalib1\n");
                        return false;
                    }
                    //Memory.CalibData1 = (Int16)(tmp & TargetDevice.CalibMask1);
                    Memory.CalibDatas[i] = (Int16)(((UInt16)tmp | (UInt16)(~TargetDevice.CalibMasks[i])) & TargetDevice.ProgMask);
                }
                else
                {   //CalibAddrs Illeagal
                    PowerDown();
                    return false;
                }
            }
            PowerDown();
            return true;
        }

        protected override bool RestoreCalib()
        {
            if (TargetDevice.CalibAddrs.Length == 0)
                return false;

            PowerON();

            int addr = -1;  // Reset vector = End of address
            for (int i = 0; i < TargetDevice.CalibAddrs.Length; i++)
            {
                int inc = TargetDevice.CalibAddrs[i] - addr;
                if (inc > 0)
                {
                    Increment(inc);
                    addr += inc;
                    if (!LoadProgMemory(Memory.CalibDatas[i]) || !SendBeginProg(WAITTYPE.WAIT_CONF))
                    {
                        PowerDown();
                        //fprintf(stderr, " [ERROR] RestoreCalib1\n");
                        return false;
                    }
                }
                else
                {   //CalibAddrs Illeagal
                    PowerDown();
                    return false;
                }
             }

            PowerDown();
            return true;
        }

        public override bool ChipErase()
        {

            if (TargetDevice.CalibAddrs.Length > 0)
            {
                if (TargetDevice.CalibAddrs[0] != 0)
                    if (!BackupCalib())
                        return false;

                Thread.Sleep(10);   // wait 10 ms bfore re-enter programming mode
            }
            else
                return false;

            PowerON();

            bool ret = false;
            switch (TargetDevice.EraseType)
            {
                case ESTYPE.BulkEraseF509:
                    ret = BulkeraseF509();
                    break;
                case ESTYPE.BulkEraseF675:
                    ret = BulkeraseF675();
                    break;
                case ESTYPE.BulkEraseF683:
                    ret = BulkeraseF683();
                    break;
                case ESTYPE.BulkEraseF84:
                    ret = BulkeraseF84();
                    break;
                case ESTYPE.BulkEraseF648A:
                    ret = BulkeraseF648A();
                    break;
                case ESTYPE.ChipEraseF88:
                    ret = ChiperaseF88();
                    break;
                case ESTYPE.ChipEraseF77:
                    ret = ChiperaseF77();
                    break;
                default:
                    //fprintf(stderr, " unknown erase type\n");
                    break;
            }

            PowerDown();

            if (TargetDevice.CalBackUp && TargetDevice.CalibAddrs.Length > 0)
            {
                if (TargetDevice.CalibAddrs[0] != 0)
                {
                    Thread.Sleep(10);   // wait 10 ms bfore re-enter programming mode
                    ret = ret && RestoreCalib();
                }
            }
            //else
            //{
            //    Thread.Sleep(10);   // wait 10 ms bfore re-enter programming mode
            //    ret = ret && ConfirmCalib();
            //}

            return ret;
        }

        public override bool ReadBuffer(PicMemory a_memory)
        {
            if (Memory == null)
                return false;

            int nSize, nMask;
            Int16[] Buffer;
            byte h, l;
            Int16 nData;
            int addr = -1;  //reste vector is config address

            PowerON();

 
            // Configuration    //reste vector is config address
            //nSize = TargetDevice.ConfDspSize;
            nMask = TargetDevice.ProgMask;//********************
            Buffer = a_memory.ConfMem;

            if (!ReadProgMemory(out nData))
            {
                PowerDown();
                return false;
            }
            Buffer[Buffer.Length - 1] = (Int16)(nData & nMask);
            Increment();    // Address = 0x0000
            addr++;

            //Progaram memory
            nSize = TargetDevice.ProgSize;
            nMask = TargetDevice.ProgMask;
            Buffer = a_memory.ProgMem;
            if (!(SendCmd(W_RECDAT_N) &&
                    SendCmd(P_READ_PRGMEM) &&
                    SendCmd((byte)(nSize & 0xff)) &&
                    SendCmd((byte)((nSize >> 8) & 0xff))))
            {
                PowerDown();
                return false;
            }
            addr += nSize;

            for (int i = 0; i < nSize; i++)
            {
                if (RecvChar(out l) && RecvChar(out h))
                {
                    Buffer[i] = (Int16)((h * 0x100 + (l & 0xff)) & nMask);
                }
                else
                {
                    //fprintf(stderr, "Error\n");
                    PowerDown();
                    return false;
                }
                //if (!m_bQuite && i%0x20 == 0)
                //    fprintf(stderr, ".");
            }

            //if (!m_bQuite)
            //fprintf(stderr, "Done\n");

            if (RecvChar(out l))//Error if extra data recieved 
            {
                //fprintf(stderr, "communication error [%02x]\n", l);
                PowerDown();
                return false;
            }

            //if (TargetDevice.DataSize > 0)
            //{
            //    //if (!m_bQuite)
            //    //    fprintf(stderr, "Data Area Reading");

            //    nSize = TargetDevice.DataSize;
            //    nMask = TargetDevice.DataMask;
            //    Buffer = a_memory.DataMem;

            //    for (int i = 0; i < nSize; i++)
            //    {
            //        if (!ReadDataMemory(out nData))
            //        {
            //            PowerDown();
            //            return false;
            //        }
            //        Buffer[i] = (Int16)(nData & nMask);
            //        Increment();

            //        //if (!m_bQuite && i%0x10 == 0)
            //        //  fprintf(stderr, ".");
            //    }

            //    //if (!m_bQuite)
            //    //    fprintf(stderr, "Done\n");
            //}

 
            //if (!m_bQuite)
            //    fprintf(stderr, "Done\n");

            PowerDown();
            return true;
        }

        public override bool WriteBuffer()
        {
            if (Memory == null)
                return false;

            PowerON();

            int nSize, nMask;
            Int16[] pBuffer;
            int addr = -1;  //reste vector is config address

            // Configuration at Last of memory
            //nMask = TargetDevice.ProgMask;//********************
            pBuffer = Memory.ConfMem;

            Int16 nConf;
            if (!ReadProgMemory(out nConf))
            {
                PowerDown();
                return false;
            }

            nConf = (Int16)((pBuffer[pBuffer.Length - 1] & TargetDevice.ConfMask[0]) | (nConf & ~TargetDevice.ConfMask[0]));
            if (!LoadProgMemory(nConf))
            {
                //fprintf (stderr, "[ERROR]: Program4\n");
                PowerDown();
                return false;
            }
            if (!SendBeginProg(WAITTYPE.WAIT_CONF))
            {
                PowerDown();
                return false;
            }
            Increment();
            addr++;


            // Program Memory
            nSize = TargetDevice.ProgSize;
            nMask = TargetDevice.ProgMask;
            pBuffer = Memory.ProgMem;
            if (TargetDevice.CalibAddrs.Length > 0)
                if ((TargetDevice.CalibAddrs[0] > 0) && (TargetDevice.CalibAddrs[0] < TargetDevice.IDAddr))
                    nSize--;

            bool bAdv15 = Advance;
            int nIncrement = 0;
            int LatSize = TargetDevice.ProgLatch <= 4 ? 4 : TargetDevice.ProgLatch;

            for (int i = 0; i < nSize; )
            {
                //if (!m_bQuite && i%0x20 == 0)
                //{
                //    fprintf(stderr, ".");
                //}

                if (pBuffer[i] != nMask)
                {
                    if (nIncrement > 0)
                    {
                        //if (nIncrement==1)
                        //    Increment();
                        //else
                        Increment(nIncrement);
                        nIncrement = 0;
                    }

                    // I don't know why condition i%8 is necessary but it needs.
                    if (bAdv15 && i % LatSize == 0 && i < nSize - LatSize)
                    {
                        //if ( LoadProgMemory8(pBuffer+i))
                        if (LoadProgMemory4N(pBuffer, i, LatSize / 4))
                        {
                            i += LatSize;
                            continue;
                        }
                        bAdv15 = false;
                    }

                    if (LoadProgMemory(pBuffer[i]))
                    {
                        if (!SendBeginProg(WAITTYPE.WAIT_PROG))
                        {
                            PowerDown();
                            return false;
                        }
                    }
                    else
                    {
                        PowerDown();
                        return false;
                    }
                }
                i++;
                nIncrement++;
            }

            // User ID Ignore

            //nSize = TargetDevice.ConfDspSize;
            //nMask = TargetDevice.ConfMask1;
            //pBuffer = Memory.ConfMem;

            ////if (!m_bQuite)
            ////    fprintf(stderr, "Config Area Writing");

            //LoadConfig(TargetDevice.ProgMask);
            //for (int i = 0; i < 4; i++)
            //{
            //    //if (!LoadProgMemory((Int16)(pBuffer[i] & 0x0f)))
            //    Int16 Uid = (Int16)(pBuffer[i] & TargetDevice.UIDMask);
            //    if (Uid != TargetDevice.UIDMask)
            //    {
            //        if (!LoadProgMemory(Uid))
            //        {
            //            //fprintf (stderr, "[ERROR]: Program1\n");
            //            PowerDown();
            //            return false;
            //        }
            //        //if (!SendBeginProg(WAITTYPE.WAIT_PROG))   // original
            //        if (!SendBeginProg(WAITTYPE.WAIT_CONF))
            //        {
            //            PowerDown();
            //            return false;
            //        }
            //    }
            //    Increment();
            //}

            PowerDown();
            return true;
        }

        public int VerifyBuffer()
        {
            // returns
            //  -1: error
            //  0: OK
            //  0b0001: NG in program memory
            //  0b0010: NG in Data memory
            //  0b0100: NG in User ID
            //  0b1000: NG in Configuration

            int nSize, nMask;
            Int16[] rBuffer, wBuffer;
            PicMemory MemoryRead = new PicMemory(TargetDevice);
            if (!ReadBuffer(MemoryRead))
                return -1;

            //if (!m_bQuite)
            //    fprintf(stderr, "Program Area Verifing");

            int ret = 0;
            nSize = TargetDevice.ProgSize;
            nMask = TargetDevice.ProgMask;
            wBuffer = Memory.ProgMem;
            rBuffer = MemoryRead.ProgMem;

            if (TargetDevice.CalibAddrs.Length > 0)
                if (TargetDevice.CalibAddrs[0] > 0 && TargetDevice.CalibAddrs[0] < TargetDevice.IDAddr)
                    nSize--;

            for (int i = 0; i < nSize; i++)
            {
                if (wBuffer[i] != rBuffer[i])
                {
                    ret += 1;
                }
            }

            // Data Memory of Base Line Family is included in program memory
            //if (TargetDevice.DataSize > 0)
            //{
            //    //if (!m_bQuite)
            //    //    fprintf(stderr, "Data Area Verifing");

            //    nSize = TargetDevice.DataSize;
            //    nMask = TargetDevice.DataMask;
            //    wBuffer = Memory.DataMem;
            //    rBuffer = MemoryRead.DataMem;

            //    for (int i = 0; i < nSize; i++)
            //    {
            //        if (wBuffer[i] != rBuffer[i])
            //        {
            //            ret += 2; ;
            //        }
            //    }
            //}

            nSize = TargetDevice.ConfDspSize;
            wBuffer = Memory.ConfMem;
            rBuffer = MemoryRead.ConfMem;
            // UserID Ignore
            //for (int i = 0; i < 4; i++)
            //{
            //    if ((rBuffer[i] & TargetDevice.UIDMask) != (wBuffer[i] & TargetDevice.UIDMask))
            //    {
            //        ret += 4; ;
            //    }
            //}

            // Configuration word
            if ((rBuffer[nSize - 1] & TargetDevice.ConfMask[0]) != (wBuffer[nSize - 1] & TargetDevice.ConfMask[0]))
            {
                ret |= 8;
            }

            return ret;
        }


    }

}
