﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Collections;

namespace W509CS
{
    enum PWTYPE
    {
        VddBeforeVpp,
        VppBeforeVdd
    };

    enum PGTYPE
    {
        Internal08,
        Internal18,
        External08,
        External18,
        External18A
    };

    enum ESTYPE
    {
        BulkEraseF509,
        BulkEraseF675,
        BulkEraseF683,
        BulkEraseF84,
        BulkEraseF648A,
        ChipEraseF88,
        ChipEraseF77
    };

    enum PICFAMILY
    {
        PICBaseLineFamily,
        PIC16Family,
        PIC16EnhFamily,
		PIC18Family,
		PIC24Family,
		dsPICFamily,
	};


    class PicDevice
    {
        //private

        public String Name { get; private set; }

        public int IDword { get; private set; }
        public int IDAddr { get; private set; }
        public PWTYPE PowerUp { get; private set; }
        public Int16 ProgMask { get; private set; }
        public int ProgSize { get; private set; }
        public int DataSize { get; private set; }
        public int DataAddr { get; private set; }
        public int ConfDspSize { get; private set; }    //???
        public int[] ConfAddr { get; private set; }
        //public int ConfAddr2 { get; private set; }
        public Int16[] ConfMask { get; private set; }
        //public Int16 ConfMask2 { get; private set; }
        //public int CalibAddr1 { get; private set; }
        //public int CalibAddr2 { get; private set; }
        //public Int16 CalibMask1 { get; private set; }
        //public Int16 CalibMask2 { get; private set; }
        public int[] CalibAddrs { get; private set; }
        public Int16[] CalibMasks { get; private set; }
        public PGTYPE ProgType { get; private set; }
        public PGTYPE ProgTypeC { get; private set; }
        public int ProgWait { get; private set; }
        public int ProgWaitD { get; private set; }
        public int ProgWaitC { get; private set; }
        public int ProgLatch { get; private set; }
        public ESTYPE EraseType { get; private set; }
        public int EraseWait { get; private set; }
        public Int16 PaMask { get; private set; }
        public bool LVP { get; private set; }
        public bool CalBackUp { get; private set; } // Need Cal back up or not //CDC original
        //additional
        //public int ProgMemAddr { get; private set; }
        //public int ConfMemAddr { get; private set; }
        //public int DataMemAddr { get; private set; }

        public PICFAMILY Family { get; private set; }
        //public Int16 ConfDspSize {get{ return 0x0f;}}


        public Int16 DataMask { get { return 0xff; } }
        public Int16 UIDMask { get; private set; }

        private PicDevice m_pDefault;

        private static ArrayList DeviceArrayList;
        public static PicDevice[] DeviceList
        {
            get
            {
                return (PicDevice[])DeviceArrayList.ToArray(typeof(PicDevice));
            }
        }

        public PicDevice(String DeviceName)
        {
            //m_pNext=null;
            Name = DeviceName;

	        IDword	   = 0;
	        IDAddr	   = 0x2000;
            UIDMask    = 0x0f;
	        PowerUp	   = PWTYPE.VddBeforeVpp;
	        ProgMask	= 0x3fff;
	        ProgSize	= 0;
	        DataSize    = 0;
            DataAddr    = 0x2100;
	        ConfDspSize = 8;
	        ConfAddr   = new int[] {0x2007};
	        //ConfAddr2   = 0;
	        ConfMask   = new Int16[] {0x3fff};
	        //ConfMask2   = 0x3fff;
            //for (int i=0;i<(sizeof(m_aCalibAddrs)/sizeof(m_aCalibAddrs[0]));i++){
            //    m_aCalibAddrs[i] = 0;
            //    m_aCalibMasks[i] = 0;
            //}
            CalibAddrs = null;
            CalibMasks = null;
	        ProgType	   = PGTYPE.Internal08;
	        ProgTypeC   =  PGTYPE.Internal08;
	        ProgWait	   = 40;
	        ProgWaitD   = 40;
	        ProgWaitC   = 40;
	        ProgLatch   = 1;
	        EraseType   = ESTYPE.BulkEraseF84;
	        EraseWait   = 40;
	        CalBackUp = true;

            if (Name.Contains("F5") || Name.StartsWith("PIC10"))
		        Family = PICFAMILY.PICBaseLineFamily;
            else if ((Name.StartsWith("PIC16") || Name.StartsWith("PIC12")) && Name.Contains("F1"))
		        Family = PICFAMILY.PIC16EnhFamily;
            else
                Family = PICFAMILY.PIC16Family;
        }

        ~PicDevice()
        {
        }
  
        //public Device m_pNext;

 
        public override string ToString(){  // "override"がないとcomboboxに表示されない 
	        return Name;
        }



        public static PicDevice Load(String Name, StreamReader sr)
        {
            String buf;
            PicDevice RetDevice = new PicDevice(Name);
            PicDevice DefDevice;
            //int i, l;
            int ConfAddr1 = 0, ConfAddr2 = 0;
            Int16 ConfMask1 = 0x3FFF, ConfMask2 = 0x3FFF;

            while (sr.Peek() >= 0)
            {
                buf = sr.ReadLine();
                //i = 0;

                //buf = buf.TrimStart(new char[] { ' ', '\t' });
                string[] Data = buf.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);

                //if (buf[i]==COMMENTCHAR)
                //    continue;

                //dbprintf(20, "Loading... %s", buf+i);

                if (Data[0] == ConfigStr.DEFENDSTR)
                    break;

                try
                {
                    System.Globalization.NumberStyles StylesHex = System.Globalization.NumberStyles.HexNumber;
                    System.Globalization.NumberStyles StylesInt = System.Globalization.NumberStyles.Integer;
                    switch (Data[0])
                    {
                        case ConfigStr.DEFAULTSTR:
                            DefDevice = SerchFromName(Data[1]);
                            if (DefDevice == null)
                                return null;

                            RetDevice = (PicDevice)DefDevice.MemberwiseClone();
                            RetDevice.Name = Name;
                            break;
                        case ConfigStr.IDWORDSTR:
                            RetDevice.IDword = int.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.IDADDRSTR:
                            RetDevice.IDAddr = int.Parse(Data[1], StylesHex);
                            //if (m_nIDAddr != 0x2000)
                            //{
                            //    fprintf(stderr, "[ERROR]: sorry but this version does not support as UserIDAddress != 0x2000\n");
                            //    return false;
                            //}
                            break;

                        case ConfigStr.UIDMASKSTR:
                            RetDevice.UIDMask = Int16.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.PWUPTYPESTR:
                            RetDevice.PowerUp = (PWTYPE)Enum.Parse(typeof(PWTYPE), Data[1]);
                            break;

                        case ConfigStr.PMMASKSTR:
                            RetDevice.ProgMask = Int16.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.PMSIZESTR:
                            RetDevice.ProgSize = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.DATASIZESTR:
                            RetDevice.DataSize = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.DATAADDRSTR:     // CDC original
                            RetDevice.DataAddr = int.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.CFGSIZESTR:
                            RetDevice.ConfDspSize = int.Parse(Data[1], StylesInt);
                            if (RetDevice.Family == PICFAMILY.PICBaseLineFamily)
                                RetDevice.ConfDspSize++;
                            break;

                        case ConfigStr.CFGADDR1STR:
                            ConfAddr1 = int.Parse(Data[1], StylesHex);
                            //if (m_nConfAddr1 != 0x2007)
                            //{
                            //    fprintf(stderr, "[ERROR]: sorry but this version does not support as ConfigAddress1 != 0x2007\n");
                            //    return false;
                            //}
                            break;

                        case ConfigStr.CFGADDR2STR:
                            ConfAddr2 = int.Parse(Data[1], StylesHex);
                            //if (!(m_nConfAddr2 == 0 || m_nConfAddr2 == 0x2008))
                            //{
                            //    fprintf(stderr, "[ERROR]: sorry but this version does not support as ConfigAddress2 != 0 or 0x2008\n");
                            //    return false;
                            //}
                            break;

                        case ConfigStr.CFGMSK1STR:
                            ConfMask1 = Int16.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.CFGMSK2STR:
                            ConfMask2 = Int16.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.CALBADDRSTR:
                            RetDevice.CalibAddrs = new int[Data.Length - 1];
                            for(int j=1; j<Data.Length; j++)
                                RetDevice.CalibAddrs[j - 1] = int.Parse(Data[j], StylesHex);
                            break;

                        case ConfigStr.CALBMASKSTR:
                            RetDevice.CalibMasks = new Int16[Data.Length - 1];
                            for(int j=1; j<Data.Length; j++)
                                RetDevice.CalibMasks[j - 1] = Int16.Parse(Data[j], StylesHex);
                            break;

 
                        case ConfigStr.PRGTYPESTR:
                            RetDevice.ProgType = (PGTYPE)Enum.Parse(typeof(PGTYPE), Data[1]);
                            break;

                        case ConfigStr.PRGTYPECSTR:
                            RetDevice.ProgTypeC = (PGTYPE)Enum.Parse(typeof(PGTYPE), Data[1]);
                            break;

                        case ConfigStr.PRGWAITSTR:
                            RetDevice.ProgWait = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.PRGWAITDSTR:
                            RetDevice.ProgWaitD = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.PRGWAITCSTR:
                            RetDevice.ProgWaitC = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.PRGLATCHSTR:
                            RetDevice.ProgLatch = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.ERSTYPESTR:
                            RetDevice.EraseType = (ESTYPE)Enum.Parse(typeof(ESTYPE), Data[1]);
                            break;

                        case ConfigStr.ERSWAITSTR:
                            RetDevice.EraseWait = int.Parse(Data[1], StylesInt);
                            break;

                        case ConfigStr.PAMASKSTR:
                            RetDevice.PaMask = Int16.Parse(Data[1], StylesHex);
                            break;

                        case ConfigStr.CALBACKUPSTR:
                            if (Data[1] == ConfigStr.NO)
                                RetDevice.CalBackUp = false;
                            else if (Data[1] == ConfigStr.YES)
                                RetDevice.CalBackUp = true;
                            break;
 
                        default:
                            return null;

                    }//end switch
                }//end try
                catch (ArgumentException e)
                {   //　String->Hex変換に失敗
                    return null;
                }
            }//end while

            //if (m_nConfAddr2 && m_nConfMask1 != m_nConfMask2)
            //{
            //    fprintf(stderr, "[ERROR]: sorry but this version does not support as ConfigMask1 != ConfigMask2 when ConfigAddress2 == 0x2008\n");
            //    return false;
            //}
            if (ConfAddr2 > 0)
            {
                RetDevice.ConfAddr = new int[2];
                RetDevice.ConfMask = new Int16[2];
                RetDevice.ConfAddr[1] = ConfAddr2;
                RetDevice.ConfMask[1] = ConfMask2;
            }
            if (ConfAddr1 > 0)
            {
                RetDevice.ConfAddr[0] = ConfAddr1;
                RetDevice.ConfMask[0] = ConfMask1;
            }
            return RetDevice;
        }//end Load()

        //private bool ParaCopy(String SrcDeviceName)
        //{
        //    PicDevice SrcDevice = SerchFromName(SrcDeviceName);
        //    if (SrcDevice != null)
        //    {
        //        m_pDefault = SrcDevice;
        //        IDword = SrcDevice.IDword;
        //        IDAddr = SrcDevice.IDAddr;
        //        UIDMask = SrcDevice.UIDMask;
        //        PowerUp = SrcDevice.PowerUp;
        //        ProgMask = SrcDevice.ProgMask;
        //        ProgSize = SrcDevice.ProgSize;
        //        DataSize = SrcDevice.DataSize;
        //        DataAddr = SrcDevice.DataAddr;
        //        ConfDspSize = SrcDevice.ConfDspSize;
        //        ConfAddr1 = SrcDevice.ConfAddr1;
        //        ConfAddr2 = SrcDevice.ConfAddr2;
        //        ConfMask1 = SrcDevice.ConfMask1;
        //        ConfMask2 = SrcDevice.ConfMask2;
        //        //for (int i = 0; i < CalibAddrs.Length; i++)
        //        //{
        //        //    CalibAddrs[i] = SrcDevice.CalibAddrs[i];
        //        //    CalibMasks[i] = SrcDevice.CalibMasks[i];
        //        //}
        //        CalibAddrs = (int[])SrcDevice.CalibAddrs.Clone();
        //        CalibMasks = (Int16[])SrcDevice.CalibMasks.Clone();
        //        ProgType = SrcDevice.ProgType;
        //        ProgTypeC = SrcDevice.ProgTypeC;
        //        ProgWait = SrcDevice.ProgWait;
        //        ProgWaitD = SrcDevice.ProgWaitD;
        //        ProgWaitC = SrcDevice.ProgWaitC;
        //        ProgLatch = SrcDevice.ProgLatch;
        //        EraseType = SrcDevice.EraseType;
        //        EraseWait = SrcDevice.EraseWait;
        //        //m_nWriteBufSize	= SrcDevice.WriteBufSize;
        //        //m_nEraseCommand	= SrcDevice.EraseCommand;
        //        //m_eEEWriteType	= SrcDevice.EEWriteType;
        //        //m_nPanelNum		= SrcDevice.PanelNum;
        //        //m_nPanelSize	= SrcDevice.PanelSize;
        //        //m_nProgWaitID	= SrcDevice.ProgWaitID;
        //        //for(int i=0;i<(sizeof(m_aConfigMask)/sizeof(m_aConfigMask[0]));i++){
        //        //    m_aConfigMask[i]  = SrcDevice.aConfigMask[i];
        //        //    m_aConfigValue[i] = SrcDevice.aConfigValue[i];
        //        //}
        //        //for(int i=0;i<(sizeof(m_WriteLatches)/sizeof(m_WriteLatches[0]));i++)
        //        //    m_WriteLatches[i] = SrcDevice.WriteLatches[i];
        //        //for(int i=0;i<(sizeof(m_NVMCONValues)/sizeof(m_NVMCONValues[0]));i++)
        //        //    m_NVMCONValues[i] = SrcDevice.NVMCONValues[i];
        //        //m_nExecCodeSize = SrcDevice.nExecCodeSize;
        //        CalBackUp = SrcDevice.CalBackUp;
        //        Family = SrcDevice.Family;
        //        return true;
        //    }
        //    else
        //        return false;

        //}

        public static bool LoadList(StreamReader sr)
        {
            DeviceArrayList = new ArrayList();

            try
            {
                while (sr.Peek() >= 0)
                {
                    string buf;
                    string[] data;
                    do
                    {
                        buf = sr.ReadLine();
                        data = buf.Split(new char[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
                    } while (sr.Peek() >= 0 && !data[0].Equals("Begin", StringComparison.OrdinalIgnoreCase));
                    if (data[0].Equals("Begin", StringComparison.OrdinalIgnoreCase) && data.Length > 1)
                    {
                        PicDevice aDevice = PicDevice.Load(data[1], sr);
                        if(aDevice == null)
                            return false;
                        DeviceArrayList.Add(aDevice);
                    }
                }
            }
            catch
            {
                return false;
            }
            finally
            {
                sr.Close();
            }

            //DeviceList = new Device[tmpList.Count];
            //tmpList.CopyTo(DeviceList);
            //DeviceList = (PicDevice[])DeviceArrayList.ToArray(typeof(PicDevice));
            return true;

        }

        public static PicDevice SerchFromName(string DeviceName)
        {
            PicDevice TheDevice = null;
            foreach(PicDevice aDevice in DeviceList)
            {
                if (DeviceName.Equals(aDevice.Name))
                {
                    TheDevice = aDevice;
                    break;
                }
            }
            return TheDevice;
        }
    }//end class Device

}
