﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.IO.Ports;

namespace W509CS
{
    delegate void DelegLog(String strLog);

    class Writer509
    {
        public DelegLog AppendLog { get; set; }

        public String PortName
        {
            get
            {
                return MyPort.PortName;
            }
            set
            {
                MyPort.PortName = value;
            }
        }

        public int BaudRate
        {
            get
            {
                return MyPort.BaudRate;
            }
            set
            {
                MyPort.BaudRate = value;
            }
        }


        public String StrVer
        {
            get
            {
                  return MyCommand.StrVer;
            }
        }

        public bool HasMemory
        {
            get
            {
                if (MyMemory == null)
                    return false;
                else
                    return true;
            }
        }

        public PicDevice TargetDevice { get; set;}
        public String StrLog { get; private set; }

        private PicCommand MyCommand;
        private PicMemory MyMemory;
        private SerialPort MyPort;
        //private PicMemory MemFromDevice;

        public Writer509()
        {
            MyCommand = new PicCommand();
            
            MyPort = new SerialPort();
            MyPort.StopBits = StopBits.One;
            MyPort.DataBits = 8;
            MyPort.Parity = Parity.None;
            MyPort.BaudRate = 38400;
            MyPort.WriteTimeout = 100;
            MyPort.ReadTimeout = 100;

            StrLog = String.Empty;
        }

        private bool PortOpen()
        {
            try
            {
                if(!MyPort.IsOpen)
                    MyPort.Open();
            }
            catch
            {
                return false;
            }
            return true;
        }

        private bool SetPicCommand()
        {
            if (TargetDevice == null)
                return false;

            if (TargetDevice.Family == PICFAMILY.PIC16Family || TargetDevice.Family == PICFAMILY.PIC16EnhFamily)
            {
                MyCommand = new PicCommand(TargetDevice, MyPort);
                return true;
            }
            else if (TargetDevice.Family == PICFAMILY.PICBaseLineFamily)
            {
                MyCommand = new PicCommandBL(TargetDevice, MyPort);
                return true;
            }
            return false;
        }

        public bool CheckVer()
        {
            MyCommand = new PicCommand();
            MyCommand.Serial = MyPort;
            if (!PortOpen())
                return false;

            bool status = MyCommand.CheckWriter();
            MyPort.Close();

            return status;
        }

        //public bool CheckChip(PicDevice TargetDevice)
        //{
        //    if (!SetPicCommand())
        //    {
        //        return false;
        //    }
        //    MyCommand.TargetDevice = TargetDevice;
        //    bool status = MyPort.Open()
        //        && MyCommand.DeviceCheck();
        //    MyPort.Close();
        //    return status;
        //}

        private bool InitialCheck()
        {
            if (!SetPicCommand())
            {
                AppendLog("Error: Illeagal device.");
                return false;
            }

            if (!PortOpen())
            {
                AppendLog("Error: Port cannot open.");
                return false;
            }

            if (!MyCommand.CheckWriter())
            {
                MyPort.Close(); 
                AppendLog("Error: No Writer was found.");
                return false;
            }
            if (!MyCommand.DeviceCheck())
            {
                MyPort.Close();
                AppendLog("Error: Mismatch Device ID.");
                return false;
            }
            return true;
        }

        public bool LoadHex(StreamReader srHexFile)
        {
            if (TargetDevice == null)
                return false;

            MyMemory = new PicMemory(TargetDevice);
            //MemToWrite.Init(TargetDevice);
            //StreamReader SrHex = new StreamReader(StrmHex, System.Text.Encoding.GetEncoding("shift_jis"));
            //StreamReader SrHex = new StreamReader(HexFile);
            bool flg = MyMemory.LoadFromFile(srHexFile);
            srHexFile.Close();

            return flg;
        }

        public bool SaveHex(StreamWriter swHexFile)
        {
            if (MyMemory == null)
                return false;

            //MyMemory = new PicMemory(TargetDevice);
            //MemToWrite.Init(TargetDevice);
            //StreamReader SrHex = new StreamReader(StrmHex, System.Text.Encoding.GetEncoding("shift_jis"));
            //StreamWriter SwHex = new StreamWriter(HexFile);
            bool status = MyMemory.saveToFile(swHexFile);
            swHexFile.Close();

            return status;
        }


        public bool WriteChip(bool DoErase)
        {
            if (TargetDevice == null || MyMemory == null)
                return false;

            if (!InitialCheck())
                return false;

            MyCommand.Memory = MyMemory;
            int status;
            try
            {
                // back up cal
                //MyCommand.Advance = false;
                if (DoErase)
                {
                    if (MyCommand.ChipErase())
                    {
                        AppendLog("Device Erased.");
                    }
                    else
                    {
                        throw new Exception("Error:Device Erase.");
                     }
                }
                AppendLog("Writeing...");
                if (MyCommand.WriteBuffer())
                {
                    AppendLog("Done.");
                }
                else
                {
                    throw new Exception("Error: Program write.");
                }
                AppendLog("Verifying...");
                status = MyCommand.VerifyBuffer();
                if (status < 0)
                {
                    throw new Exception("Error: Verify error.");
                }
 
            }
            catch(Exception e)
            {
                AppendLog(e.Message);
                return false;
            }
            finally
            {
                if(MyPort.IsOpen)
                    MyPort.Close();
            }
            if (status == 0)
            {
                AppendLog("Verify OK.");
                return true;
            }
            else
            {
                AppendLog("Verify Fallure.");
                if ((status & 1) == 1)
                    AppendLog("Program Memory Missmatched.");
                if ((status & 2) == 2)
                    AppendLog("Data Memory Missmatched.");
                if ((status & 4) == 4)
                    AppendLog("User ID Missmatched.");
                if ((status & 8) == 8)
                    AppendLog("Configuration Missmatched.");
                return false; ;
            }
        }

        public bool ReadFromChip()
        {
            if (TargetDevice == null)
                return false;


            if (!InitialCheck())
                return false;

            MyMemory = new PicMemory(TargetDevice);
            MyCommand.Memory = MyMemory;

            AppendLog("Reading...");

            bool status = MyCommand.ReadBuffer();
            MyPort.Close();

            return status;
        }

        public bool EraseChip()
        {
            if (TargetDevice == null)
                return false;

            // check device
            if (!InitialCheck())
                return false;


            // back up cal
            //MyCommand.Advance = false;
            bool status = MyCommand.ChipErase();

            MyPort.Close();
            return status;
        }


        public bool VerifyChip(bool isBlank)
        {
            // bool isBlank:    true = check blank
            //                  false = verify

            if (TargetDevice == null)
                return false;

            // check writer & device
            if (!InitialCheck())
                return false;

            if (isBlank)
            {
                MyCommand.Memory = new PicMemory(TargetDevice);
            }
            else
            {
                if (MyMemory == null)
                {
                    MyPort.Close();
                    return false;
                }
                MyCommand.Memory = MyMemory;
            }

            AppendLog("Reading...");

            // back up cal
            //MyCommand.Advance = false;
            int status = MyCommand.VerifyBuffer();

            MyPort.Close();

            if (status == 0)
            {
                 return true;
            }
            else
            {
                if ((status & 1) == 1)
                    AppendLog("Program Memory: Wrong");
                if ((status & 2) == 2)
                    AppendLog("Data Memory: Wrong");
                if ((status & 4) == 4)
                    AppendLog("User ID: Wrong");
                if ((status & 8) == 8)
                    AppendLog("Configuration: Wrong");
                return false; ;
            }
        }

        public String MemoryDump()
        {
            if(MyMemory == null)
            {
                return String.Empty;
            }
            String sDump = "Code:" + Environment.NewLine;

            int Size = TargetDevice.ProgSize;
            int Mask = TargetDevice.ProgMask;
            Int16[] Buf = MyMemory.ProgMem;
            //if(TargetDevice.CalibAddrs.Length > 0)
            //    if(TargetDevice.CalibAddrs[0] > 0 && TargetDevice.CalibAddrs[0] < 0x2000)
            //        Size--;
            for(int i= 0; i < Size; )
            {
                sDump += i.ToString("X4") + ":\t";
                for (int j = 0; j < 8 && i < Size; j++)
                {
                    int Val = Buf[i++] & Mask;
                    sDump += Val.ToString("X4") + "  ";
                }
                sDump += Environment.NewLine;
            }

            //Data
            sDump += Environment.NewLine;
            sDump += "Data:" + Environment.NewLine;

            Size = TargetDevice.DataSize;
            Mask = TargetDevice.DataMask;
            Buf = MyMemory.DataMem;
            if(Size > 0)
            {
                for(int i= 0; i < Size; )
                {
                    sDump += i.ToString("X4") + ":\t";
                    for(int j=0; j<8 && i < Size; j++)
                    {
                        int Val = Buf[i++] & Mask;
                        sDump += Val.ToString("X2") + "  ";
                    }
                    sDump += Environment.NewLine;
                }
            }
            //ID & Configuration
            sDump += Environment.NewLine;
            sDump += "ID & Configuration:" + Environment.NewLine;

            Size = TargetDevice.ConfDspSize;
            Mask = TargetDevice.ProgMask;
            Buf = MyMemory.ConfMem;
            for(int i= 0; i < Size; )
            {
                sDump += i.ToString("X4") + ":\t";
                for(int j=0; j<8 && i < Size; j++)
                {
                    int Val = Buf[i++] & Mask;
                    sDump += Val.ToString("X4") + "  ";
                }
                sDump += Environment.NewLine;
            }
            return sDump;

        }

    }
}
